/* File:        flrassert.P
**    	        trailer appended automatically by the FLORA compiler
**
** Author(s): Guizhen Yang 
** Contact:   xsb-contact@cs.sunysb.edu
**
** Copyright (C) The Research Foundation of SUNY, 1999
**
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
** 
** $Id: flrassert.P,v 1.6 1999/11/21 05:52:38 kifer Exp $
** 
*/


:- compiler_options([xpp_on]).

#define  VAR_FPREFIX_BASE  dyn_$_$_flora_
#include "flora_prefix.h"


:- import var/1, call/1 from standard.

:- import clause/2 from assert.

:- import trie_assert/1, trie_retract_nr/1 from tables.


/*****************************************************************************
  rules for assert/retract
*****************************************************************************/
STD_FPREFIX(assert)([]).

STD_FPREFIX(assert)([H|T]) :-
	trie_assert(H),
	STD_FPREFIX(assert)(T).


/****************************************************************************/
STD_FPREFIX(retract)(Gs,Cs) :-
	STD_FPREFIX(prove_goals)(Gs,Fs),
	STD_FPREFIX(prove_conditions)(Cs),
	STD_FPREFIX(check_facts)(Fs),
	STD_FPREFIX(retract_facts)(Fs).


STD_FPREFIX(prove_goals)([],[]).

STD_FPREFIX(prove_goals)([flora_rtrct(G)|Gs],[G|Fs]) :-
	!,
	call(G),
	STD_FPREFIX(prove_goals)(Gs,Fs).

STD_FPREFIX(prove_goals)([G|Gs],Fs) :-
	call(G),
	STD_FPREFIX(prove_goals)(Gs,Fs).


STD_FPREFIX(prove_conditions)([]).

STD_FPREFIX(prove_conditions)([C|Cs]) :-
	call(C),
	!,
	STD_FPREFIX(prove_conditions)(Cs).


STD_FPREFIX(check_facts)([]).

STD_FPREFIX(check_facts)([H|Fs]) :-
	clause(H,true),
	STD_FPREFIX(check_facts)(Fs).


STD_FPREFIX(retract_facts)([]).

STD_FPREFIX(retract_facts)([H|Fs]) :-
	trie_retract_nr((H :- true)),
	STD_FPREFIX(retract_facts)(Fs).


/****************************************************************************/
STD_FPREFIX(retractall)(Gs,Cs) :-
	STD_FPREFIX(retract)(Gs,Cs),
	fail.

STD_FPREFIX(retractall)(_,_).


/****************************************************************************/
STD_FPREFIX(erase)(Gs,Cs) :-
	STD_FPREFIX(prove_goals)(Gs,Fs),
	STD_FPREFIX(prove_conditions)(Cs),
	STD_FPREFIX(check_facts)(Fs),
	STD_FPREFIX(erase_facts)(Fs).


STD_FPREFIX(erase_facts)([]).

STD_FPREFIX(erase_facts)([H|Fs]) :-
	STD_FPREFIX(erase_one_fact)(H),
	STD_FPREFIX(erase_facts)(Fs).


STD_FPREFIX(erase_one_fact)(G) :-
	trie_retract_nr(G),
	( G = VAR_FPREFIX(exists)(O)     -> true
	;   G = VAR_FPREFIX(isa)(O,_)    -> true
	;   G = VAR_FPREFIX(fd)(_,_,O)   -> true
	;   G = VAR_FPREFIX(mvd)(_,_,O)  -> true
	;   G = VAR_FPREFIX(ifd)(_,_,O)  -> true
	;   G = VAR_FPREFIX(imvd)(_,_,O) -> true
	;   G = VAR_FPREFIX(fs)(_,_,O)   -> true
	;   G = VAR_FPREFIX(mvs)(_,_,O)  -> true
	;   NextAction=no
	),
	( NextAction==no ->
	    true

	  ; var(O) ->
	    true
	  ;
	    STD_FPREFIX(erase_object)(O)
	).


STD_FPREFIX(erase_object)(O) :-
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(exists)(O))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(isa)(O,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(sub)(O,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(fd)(O,_,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(mvd)(O,_,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(mvd)(O,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(ifd)(O,_,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(imvd)(O,_,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(imvd)(O,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(fs)(O,_,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(mvs)(O,_,_))],[]),
	STD_FPREFIX(eraseall)([flora_rtrct(VAR_FPREFIX(mvs)(O,_))],[]).


/****************************************************************************/
STD_FPREFIX(eraseall)(Gs,Cs) :-
	STD_FPREFIX(erase)(Gs,Cs),
	fail.

STD_FPREFIX(eraseall)(_,_).


#undef VAR_FPREFIX_BASE

/****************************  End of flrassert.P  *******************/

